# Generated by Django 5.2.1 on 2025-05-19 15:01
from typing import TYPE_CHECKING

import django.core.validators
import django.db.migrations.operations.special
import django.db.models.deletion
import taggit.managers
import utilities.json
from django.conf import settings
from django.db import migrations
from django.db import models
from extras.choices import CustomFieldTypeChoices
from extras.choices import CustomFieldUIEditableChoices
from extras.choices import CustomFieldUIVisibleChoices
from extras.choices import CustomLinkButtonClassChoices

from ipfabric_netbox.utilities.transform_map import build_transform_maps
from ipfabric_netbox.utilities.transform_map import get_transform_map

if TYPE_CHECKING:
    from django.apps import apps as apps_type
    from django.db.backends.base.schema import BaseDatabaseSchemaEditor

# This is squashed migration since when StagedChanges were removed, NetBox Branch model
# was moved to netbox branching. But IPFabricBranch was extending it and I had to remove it.
# We need to avoid creating IPFabricBranch on NetBox >4.3.0 since Branch is not there.
# So instead the squash does not create it and conditionally removed the table.


def create_custom_field(
    apps: "apps_type",
    db_alias: str,
    field_name: str,
    label: str,
    target_models: list,
    object_type=None,
    cf_type: str | None = "type_text",
):
    """Create a single custom field and link it to required models."""
    ObjectType = apps.get_model("core", "ObjectType")
    ContentType = apps.get_model("contenttypes", "ContentType")

    defaults = {
        "label": label,
        "related_object_type": (
            ObjectType.objects.get_for_model(object_type) if object_type else None
        ),
        "ui_visible": getattr(CustomFieldUIVisibleChoices, "ALWAYS"),
        "ui_editable": getattr(CustomFieldUIEditableChoices, "NO"),
    }

    custom_field, _ = (
        apps.get_model("extras", "CustomField")
        .objects.using(db_alias)
        .update_or_create(
            type=getattr(CustomFieldTypeChoices, cf_type.upper()),
            name=field_name,
            defaults=defaults,
        )
    )

    for model in target_models:
        custom_field.object_types.add(ContentType.objects.get_for_model(model))


def prepare_custom_fields(apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"):
    """Forward migration to prepare ipfabric_netbox custom fields and links."""
    db_alias = schema_editor.connection.alias
    Device = apps.get_model("dcim", "Device")
    Site = apps.get_model("dcim", "Site")

    create_custom_field(
        apps,
        db_alias,
        "ipfabric_source",
        "IP Fabric Source",
        [Device, Site],
        cf_type="type_object",
        object_type=apps.get_model("ipfabric_netbox", "IPFabricSource"),
    )
    create_custom_field(
        apps,
        db_alias,
        "ipfabric_ingestion",
        "IP Fabric Last Ingestion",
        [Device, Site],
        cf_type="type_object",
        object_type=apps.get_model("ipfabric_netbox", "IPFabricIngestion"),
    )
    cl, _ = (
        apps.get_model("extras", "CustomLink")
        .objects.using(db_alias)
        .update_or_create(
            defaults={
                "link_text": "{% if object.custom_field_data.ipfabric_source is defined %}{% set SOURCE_ID = object.custom_field_data.ipfabric_source %}{% if SOURCE_ID %}IP Fabric{% endif %}{% endif %}",
                "link_url": '{% if object.custom_field_data.ipfabric_source is defined %}{% set SOURCE_ID = object.custom_field_data.ipfabric_source %}{% if SOURCE_ID %}{% set BASE_URL = object.custom_fields.filter(related_object_type__model="ipfabricsource").first().related_object_type.model_class().objects.get(pk=SOURCE_ID).url %}{{ BASE_URL }}/inventory/devices?options={"filters":{"sn": ["like","{{ object.serial }}"]}}{% endif %}{%endif%}',
                "new_window": True,
                "button_class": CustomLinkButtonClassChoices.BLUE,
            },
            name="ipfabric",
        )
    )
    cl.object_types.add(
        apps.get_model("contenttypes", "ContentType")
        .objects.db_manager(db_alias)
        .get_for_model(Device)
    )


def cleanup_custom_fields(apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"):
    """Reverse migration to prepare ipfabric_netbox custom fields and links."""
    db_alias = schema_editor.connection.alias
    for field_name in ["ipfabric_source", "ipfabric_ingestion"]:
        custom_field = (
            apps.get_model("extras", "CustomField")
            .objects.using(db_alias)
            .get(name=field_name)
        )
        for model in custom_field.object_types.all()[:]:
            custom_field.object_types.remove(model)
        custom_field.delete()


def prepare_transform_maps(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Create transform maps if they do not exist yet.
    They used to be created during plugin.ready() so they might be present on older DBs.
    """
    build_transform_maps(
        data=get_transform_map(), apps=apps, db_alias=schema_editor.connection.alias
    )


class Migration(migrations.Migration):
    replaces = [
        ("ipfabric_netbox", "0001_initial"),
        ("ipfabric_netbox", "0002_ipfabricsnapshot_status"),
        ("ipfabric_netbox", "0003_ipfabricsource_type_and_more"),
        ("ipfabric_netbox", "0004_ipfabricsync_auto_merge"),
        (
            "ipfabric_netbox",
            "0005_alter_ipfabricrelationshipfield_source_model_and_more",
        ),
        ("ipfabric_netbox", "0006_alter_ipfabrictransformmap_target_model"),
        ("ipfabric_netbox", "0007_prepare_custom_fields"),
        ("ipfabric_netbox", "0008_prepare_transform_maps"),
        ("ipfabric_netbox", "0009_transformmap_changes_for_netbox_v4_2"),
        ("ipfabric_netbox", "0010_remove_uuid_from_get_or_create"),
        ("ipfabric_netbox", "0011_update_part_number_DCIM_inventory_item_template"),
        ("ipfabric_netbox", "0012_remove_status_field"),
        ("ipfabric_netbox", "0013_switch_to_branching_plugin"),
    ]

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("core", "0012_job_object_type_optional"),
        ("dcim", "0191_module_bay_rebuild"),
        ("extras", "0123_journalentry_kind_default"),
        ("ipam", "0070_vlangroup_vlan_id_ranges"),
        ("netbox_branching", "0003_rename_indexes"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="IPFabricSource",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=utilities.json.CustomFieldJSONEncoder,
                    ),
                ),
                ("description", models.CharField(blank=True, max_length=200)),
                ("comments", models.TextField(blank=True)),
                ("name", models.CharField(max_length=100, unique=True)),
                ("url", models.CharField(max_length=200)),
                (
                    "status",
                    models.CharField(default="new", editable=False, max_length=50),
                ),
                ("parameters", models.JSONField(blank=True, null=True)),
                ("last_synced", models.DateTimeField(blank=True, null=True)),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
                ("type", models.CharField(default="local", max_length=50)),
            ],
            options={
                "verbose_name": "IP Fabric Source",
                "verbose_name_plural": "IP Fabric Sources",
                "ordering": ("name",),
            },
        ),
        migrations.CreateModel(
            name="IPFabricSnapshot",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("last_updated", models.DateTimeField(editable=False)),
                ("name", models.CharField(max_length=200)),
                (
                    "source",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="snapshots",
                        to="ipfabric_netbox.ipfabricsource",
                    ),
                ),
                ("snapshot_id", models.CharField(max_length=100)),
                ("status", models.CharField(default="unloaded", max_length=50)),
                ("data", models.JSONField(blank=True, null=True)),
                ("date", models.DateTimeField(blank=True, editable=False, null=True)),
            ],
            options={
                "verbose_name": "IP Fabric Snapshot",
                "verbose_name_plural": "IP Fabric Snapshots",
                "ordering": ("source", "-date"),
            },
        ),
        migrations.CreateModel(
            name="IPFabricTransformMap",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=utilities.json.CustomFieldJSONEncoder,
                    ),
                ),
                ("name", models.CharField(max_length=100, unique=True)),
                ("source_model", models.CharField(max_length=50)),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
                (
                    "target_model",
                    models.ForeignKey(
                        limit_choices_to=models.Q(
                            models.Q(
                                models.Q(("app_label", "dcim"), ("model", "site")),
                                models.Q(
                                    ("app_label", "dcim"), ("model", "manufacturer")
                                ),
                                models.Q(("app_label", "dcim"), ("model", "platform")),
                                models.Q(
                                    ("app_label", "dcim"), ("model", "devicerole")
                                ),
                                models.Q(
                                    ("app_label", "dcim"), ("model", "devicetype")
                                ),
                                models.Q(("app_label", "dcim"), ("model", "device")),
                                models.Q(
                                    ("app_label", "dcim"), ("model", "virtualchassis")
                                ),
                                models.Q(("app_label", "dcim"), ("model", "interface")),
                                models.Q(
                                    ("app_label", "dcim"), ("model", "macaddress")
                                ),
                                models.Q(("app_label", "ipam"), ("model", "vlan")),
                                models.Q(("app_label", "ipam"), ("model", "vrf")),
                                models.Q(("app_label", "ipam"), ("model", "prefix")),
                                models.Q(("app_label", "ipam"), ("model", "ipaddress")),
                                models.Q(
                                    ("app_label", "contenttypes"),
                                    ("model", "contenttype"),
                                ),
                                models.Q(("app_label", "tenancy"), ("model", "tenant")),
                                models.Q(
                                    ("app_label", "dcim"), ("model", "inventoryitem")
                                ),
                                _connector="OR",
                            )
                        ),
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "IP Fabric Transform Map",
                "verbose_name_plural": "IP Fabric Transform Maps",
            },
        ),
        migrations.CreateModel(
            name="IPFabricTransformField",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("source_field", models.CharField(max_length=100)),
                ("target_field", models.CharField(max_length=100)),
                ("coalesce", models.BooleanField(default=False)),
                ("template", models.TextField(blank=True, default="")),
                (
                    "transform_map",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="field_maps",
                        to="ipfabric_netbox.ipfabrictransformmap",
                    ),
                ),
            ],
            options={
                "verbose_name": "IP Fabric Transform Field",
                "verbose_name_plural": "IP Fabric Transform Fields",
                "ordering": ("transform_map",),
            },
        ),
        migrations.CreateModel(
            name="IPFabricSync",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                ("name", models.CharField(max_length=100, unique=True)),
                ("type", models.CharField(default="dcim", max_length=50)),
                (
                    "status",
                    models.CharField(default="new", editable=False, max_length=50),
                ),
                ("parameters", models.JSONField(blank=True, null=True)),
                (
                    "last_synced",
                    models.DateTimeField(blank=True, editable=False, null=True),
                ),
                ("scheduled", models.DateTimeField(blank=True, null=True)),
                (
                    "interval",
                    models.PositiveIntegerField(
                        blank=True,
                        null=True,
                        validators=[django.core.validators.MinValueValidator(1)],
                    ),
                ),
                (
                    "snapshot_data",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="snapshots",
                        to="ipfabric_netbox.ipfabricsnapshot",
                    ),
                ),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("auto_merge", models.BooleanField(default=False)),
            ],
            options={
                "verbose_name": "IP Fabric Sync",
                "ordering": ["pk"],
            },
        ),
        migrations.CreateModel(
            name="IPFabricData",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("data", models.JSONField(blank=True, null=True)),
                ("type", models.CharField(max_length=50)),
                (
                    "snapshot_data",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ipf_data",
                        to="ipfabric_netbox.ipfabricsnapshot",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="IPFabricRelationshipField",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("target_field", models.CharField(max_length=100)),
                ("coalesce", models.BooleanField(default=False)),
                ("template", models.TextField(blank=True, default="")),
                (
                    "source_model",
                    models.ForeignKey(
                        limit_choices_to=models.Q(
                            models.Q(
                                ("app_label", "dcim"),
                                ("app_label", "ipam"),
                                ("app_label", "tenancy"),
                                models.Q(
                                    ("app_label", "contenttypes"),
                                    ("model", "contenttype"),
                                ),
                                _connector="OR",
                            )
                        ),
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="ipfabric_transform_fields",
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "transform_map",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="relationship_maps",
                        to="ipfabric_netbox.ipfabrictransformmap",
                    ),
                ),
            ],
            options={
                "verbose_name": "IP Fabric Relationship Field",
                "verbose_name_plural": "IP Fabric Relationship Fields",
                "ordering": ("transform_map",),
            },
        ),
        migrations.CreateModel(
            name="IPFabricIngestion",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                (
                    "branch",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="netbox_branching.branch",
                    ),
                ),
                (
                    "job",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="core.job",
                    ),
                ),
                (
                    "sync",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="ipfabric_netbox.ipfabricsync",
                    ),
                ),
            ],
            options={
                "verbose_name": "IP Fabric Ingestion",
                "verbose_name_plural": "IP Fabric Ingestion",
                "ordering": ("pk",),
            },
        ),
        migrations.RunPython(
            code=prepare_custom_fields,
            reverse_code=cleanup_custom_fields,
        ),
        migrations.RunPython(
            code=prepare_transform_maps,
            reverse_code=migrations.RunPython.noop,
        ),
    ]



import importlib.resources
import json
import logging
from typing import Any
from typing import Callable

from django.apps import apps as django_apps


logger = logging.getLogger("ipfabric_netbox.utilities.transform_map")

# region Transform Map Creation

# These functions are used in the migration file to prepare the transform maps
# Because of this we have to use historical models
# see https://docs.djangoproject.com/en/5.1/topics/migrations/#historical-models


def build_fields(data, apps, db_alias):
    ContentType = apps.get_model("contenttypes", "ContentType")
    if "target_model" in data:
        ct = ContentType.objects.db_manager(db_alias).get_for_model(
            apps.get_model(
                data["target_model"]["app_label"],
                data["target_model"]["model"],
            )
        )
        data["target_model"] = ct
    elif "source_model" in data:
        ct = ContentType.objects.db_manager(db_alias).get_for_model(
            apps.get_model(
                data["source_model"]["app_label"],
                data["source_model"]["model"],
            )
        )
        data["source_model"] = ct
    return data


def build_transform_maps(data, apps: django_apps = None, db_alias: str = "default"):
    apps = apps or django_apps
    ContentType = apps.get_model("contenttypes", "ContentType")
    IPFabricTransformMap = apps.get_model("ipfabric_netbox", "IPFabricTransformMap")
    IPFabricTransformField = apps.get_model("ipfabric_netbox", "IPFabricTransformField")
    IPFabricRelationshipField = apps.get_model(
        "ipfabric_netbox", "IPFabricRelationshipField"
    )

    # Mapping for backward compatibility (endpoint -> source_model)
    # Used only when running against old model that still has source_model field
    # TODO: Remove once migrations are squashed and old versions are no longer supported
    endpoint_to_source_model = {
        "/technology/addressing/managed-ip/ipv4": "ipaddress",
        "/inventory/devices": "device",
        "/inventory/sites/overview": "site",
        "/inventory/interfaces": "interface",
        "/inventory/part-numbers": "part_number",
        "/technology/vlans/site-summary": "vlan",
        "/technology/routing/vrf/detail": "vrf",
        "/technology/networks/managed-networks": "prefix",
        "/technology/platforms/stack1/members": "virtualchassis",
    }

    model_fields = {f.name for f in IPFabricTransformMap._meta.get_fields()}
    for tm in data:
        field_data = build_fields(tm["data"], apps, db_alias)

        endpoint_value = field_data.pop("source_endpoint")

        if "source_endpoint" in model_fields:
            # New model: use source_endpoint foreign key to IPFabricEndpoint
            # This models does not exist when TMs are populated, so need to get it here
            IPFabricEndpoint = apps.get_model("ipfabric_netbox", "IPFabricEndpoint")
            try:
                endpoint = IPFabricEndpoint.objects.using(db_alias).get(
                    endpoint=endpoint_value
                )
                field_data["source_endpoint"] = endpoint
            except IPFabricEndpoint.DoesNotExist:
                # Use first endpoint as fallback
                endpoint = IPFabricEndpoint.objects.using(db_alias).first()
                field_data["source_endpoint"] = endpoint
                field_data[
                    "name"
                ] = f"[NEEDS CORRECTION - Expected endpoint '{endpoint_value}' not found] {field_data['name']}"

        else:
            # Old model: convert source_endpoint value to source_model string
            source_model_value = endpoint_to_source_model.get(endpoint_value, "device")
            field_data["source_model"] = source_model_value

        # This field was not present in the old model, so remove it if exists
        tm_parents = []
        if parents := field_data.pop("parents", None):
            if isinstance(parents, str):
                parents = [parents]
            for parent in parents:
                if "parents" not in model_fields:
                    continue
                # New model: set parents MTM field to IPFabricTransformMap
                app, model = parent.split(".")
                try:
                    parent_tm = IPFabricTransformMap.objects.using(db_alias).get(
                        target_model=ContentType.objects.using(db_alias).get(
                            app_label=app, model=model
                        ),
                        group__isnull=True,
                    )
                    tm_parents.append(parent_tm)
                except IPFabricTransformMap.DoesNotExist:
                    raise ValueError(f"Parent Transform Map '{parent}' not found")

        tm_obj = IPFabricTransformMap.objects.using(db_alias).create(**field_data)
        # Old migrations may not have parents field
        if hasattr(tm_obj, "parents"):
            tm_obj.parents.set(tm_parents)
        for fm in tm["field_maps"]:
            field_data = build_fields(fm, apps, db_alias)
            IPFabricTransformField.objects.using(db_alias).create(
                transform_map=tm_obj, **field_data
            )
        for rm in tm["relationship_maps"]:
            relationship_data = build_fields(rm, apps, db_alias)
            IPFabricRelationshipField.objects.using(db_alias).create(
                transform_map=tm_obj, **relationship_data
            )


def get_transform_map() -> dict:
    for data_file in importlib.resources.files("ipfabric_netbox.data").iterdir():
        if data_file.name != "transform_map.json":
            continue
        with open(data_file, "rb") as data_file:
            return json.load(data_file)
    raise FileNotFoundError("'transform_map.json' not found in installed package")


# endregion
# region Transform Map Updating


class Record:
    """Base class for field and relationship records."""

    def __init__(
        self,
        coalesce: bool | None = None,
        old_template: str = None,
        new_template: str = None,
    ):
        self.coalesce = coalesce
        # Keep the original template here rather than loading it from transform_map.json
        # so our revert won’t break if that template ever changes.
        self.old_template = old_template
        self.new_template = new_template


class FieldRecord(Record):
    def __init__(
        self,
        source_field: str,
        target_field: str,
        new_source_field: str | None = None,
        new_target_field: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.source_field = source_field
        self.target_field = target_field
        self.new_source_field = new_source_field
        self.new_target_field = new_target_field


class RelationshipRecord(Record):
    def __init__(self, source_model: str, target_field: str, **kwargs):
        super().__init__(**kwargs)
        self.source_model = source_model
        self.target_field = target_field


class TransformMapRecord:
    def __init__(
        self,
        target_model: str,
        fields: tuple[FieldRecord, ...] = tuple(),
        relationships: tuple[RelationshipRecord, ...] = tuple(),
        # Support both source_model string and source_endpoint string for backward compatibility
        source_model: str | None = None,
        source_endpoint: str | None = None,
    ):
        self.source_model = source_model
        self.source_endpoint = source_endpoint
        self.target_model = target_model
        self.fields = fields
        self.relationships = relationships


def do_change(
    apps, schema_editor, changes: tuple[TransformMapRecord, ...], forward: bool = True
):
    """Apply the changes, `forward` determines direction."""

    ContentType = apps.get_model("contenttypes", "ContentType")
    IPFabricTransformMap = apps.get_model("ipfabric_netbox", "IPFabricTransformMap")
    IPFabricTransformField = apps.get_model("ipfabric_netbox", "IPFabricTransformField")
    IPFabricRelationshipField = apps.get_model(
        "ipfabric_netbox", "IPFabricRelationshipField"
    )

    try:
        for change in changes:
            app, model = change.target_model.split(".")
            try:
                if change.source_model:
                    transform_map = IPFabricTransformMap.objects.get(
                        source_model=change.source_model,
                        target_model=ContentType.objects.get(
                            app_label=app, model=model
                        ),
                    )
                else:
                    transform_map = IPFabricTransformMap.objects.get(
                        source_endpoint__endpoint=change.source_endpoint,
                        target_model=ContentType.objects.get(
                            app_label=app, model=model
                        ),
                    )
            except IPFabricTransformMap.DoesNotExist:
                continue

            for field in change.fields:
                # Find the correct transform field.
                # Only 1 should be found if it exists, but keep it as queryset so we can filter and update.
                transform_field_qs = IPFabricTransformField.objects.filter(
                    transform_map=transform_map,
                    source_field=field.source_field
                    if forward
                    else field.new_source_field or field.source_field,
                    target_field=field.target_field
                    if forward
                    else field.new_target_field or field.target_field,
                )
                if not transform_field_qs.exists():
                    continue

                if field.old_template is not None and field.new_template is not None:
                    # First update the template if needed
                    transform_field_qs.filter(
                        template=field.old_template if forward else field.new_template
                    ).update(
                        template=field.new_template if forward else field.old_template
                    )

                if field.coalesce is not None:
                    # Next update coalesce if needed
                    transform_field_qs.filter(
                        coalesce=not field.coalesce if forward else field.coalesce
                    ).update(coalesce=field.coalesce if forward else not field.coalesce)

                if (
                    field.new_target_field is not None
                    or field.new_source_field is not None
                ):
                    # And at the end update source_field/target_field if needed
                    transform_field_qs.update(
                        source_field=field.new_source_field or field.source_field
                        if forward
                        else field.source_field,
                        target_field=field.new_target_field or field.target_field
                        if forward
                        else field.target_field,
                    )

            for relationship in change.relationships:
                s_app, s_model = relationship.source_model.split(".")
                source_model = ContentType.objects.get(app_label=s_app, model=s_model)

                # Find the correct relationship field.
                # Only 1 should be found if it exists, but keep it as queryset so we can filter and update.
                relationship_qs = IPFabricRelationshipField.objects.filter(
                    transform_map=transform_map,
                    source_model=source_model,
                    target_field=relationship.target_field,
                )
                if not relationship_qs.exists():
                    continue

                if (
                    relationship.old_template is not None
                    and relationship.new_template is not None
                ):
                    # First update the template if needed
                    relationship_qs.filter(
                        template=relationship.old_template
                        if forward
                        else relationship.new_template,
                    ).update(
                        template=relationship.new_template
                        if forward
                        else relationship.old_template
                    ),

                if relationship.coalesce is not None:
                    # Next update coalesce if needed
                    relationship_qs.filter(
                        coalesce=not relationship.coalesce
                        if forward
                        else relationship.coalesce,
                    ).update(
                        coalesce=relationship.coalesce
                        if forward
                        else not relationship.coalesce
                    ),

    except Exception as e:
        print(f"Error applying Transform map updates: {e}")


# endregion

# region Cycle Detection


def has_cycle_dfs(
    node_id: int,
    get_parents_func: Callable[[int, Any], Any],
    parent_override: Any = None,
    visited: set[int] | None = None,
    rec_stack: set[int] | None = None,
) -> bool:
    """
    DFS helper to detect cycles in transform map parent relationships.

    Args:
        node_id: The ID of the current node being checked
        get_parents_func: Function that takes (node_id, parent_override) and returns parent objects
        parent_override: Optional override for parents of a specific node (used for validation)
        visited: Set of already visited node IDs (created automatically if not provided)
        rec_stack: Set of node IDs in the current recursion stack (created automatically if not provided)

    Returns:
        True if a cycle is detected, False otherwise
    """
    if visited is None:
        visited = set()
    if rec_stack is None:
        rec_stack = set()

    visited.add(node_id)
    rec_stack.add(node_id)

    try:
        parents = get_parents_func(node_id, parent_override)

        for parent in parents:
            parent_pk = parent.pk if hasattr(parent, "pk") else parent.id
            if parent_pk not in visited:
                if has_cycle_dfs(
                    parent_pk, get_parents_func, visited=visited, rec_stack=rec_stack
                ):
                    return True
            elif parent_pk in rec_stack:
                # Found a back edge - cycle detected
                return True
    except Exception as err:
        logger.warning(f"Error applying Transform map updates: {err}")

    rec_stack.remove(node_id)
    return False


# endregion



from typing import TYPE_CHECKING

from django.db import migrations

from ipfabric_netbox.utilities.endpoint import build_endpoints
from ipfabric_netbox.utilities.endpoint import get_endpoint_data
from ipfabric_netbox.utilities.filters import build_filters
from ipfabric_netbox.utilities.filters import get_filter_data
from ipfabric_netbox.utilities.transform_map import get_transform_map

if TYPE_CHECKING:
    from django.apps import apps as apps_type
    from django.db.backends.base.schema import BaseDatabaseSchemaEditor


def prepare_endpoints(apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"):
    """Create endpoints if they do not exist yet."""
    build_endpoints(
        data=get_endpoint_data(), apps=apps, db_alias=schema_editor.connection.alias
    )


def prepare_filters(apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"):
    """Create filters if they do not exist yet."""
    build_filters(
        data=get_filter_data(), apps=apps, db_alias=schema_editor.connection.alias
    )


def assign_filters_to_syncs(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Assign all created IPFabricFilters to all existing IPFabricSync objects."""
    IPFabricSync = apps.get_model("ipfabric_netbox", "IPFabricSync")
    IPFabricFilter = apps.get_model("ipfabric_netbox", "IPFabricFilter")

    all_filters = IPFabricFilter.objects.using(schema_editor.connection.alias).all()
    for sync in IPFabricSync.objects.using(schema_editor.connection.alias).all():
        sync.filters.set(all_filters)


def migrate_source_model_to_endpoint(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Migrate IPFabricTransformMap source_model data to source_endpoint."""
    IPFabricTransformMap = apps.get_model("ipfabric_netbox", "IPFabricTransformMap")
    IPFabricEndpoint = apps.get_model("ipfabric_netbox", "IPFabricEndpoint")
    ContentType = apps.get_model("contenttypes", "ContentType")

    source_model_to_endpoint = {
        "site": "/inventory/sites/overview",
        "device": "/inventory/devices",
        "virtualchassis": "/technology/platforms/stack/members",
        "interface": "/inventory/interfaces",
        "part_number": "/inventory/part-numbers",
        "vlan": "/technology/vlans/site-summary",
        "vrf": "/technology/routing/vrf/detail",
        "prefix": "/technology/networks/managed-networks",
        "ipaddress": "/technology/addressing/managed-ip/ipv4",
        "inventory": "/inventory/part-numbers",
    }

    # Get first endpoint as fallback to make sure it's always set
    fallback_endpoint = IPFabricEndpoint.objects.using(
        schema_editor.connection.alias
    ).first()

    for transform_map in IPFabricTransformMap.objects.using(
        schema_editor.connection.alias
    ).all():
        endpoint_value = source_model_to_endpoint.get(transform_map.source_model)

        # Special case: source_model="device" with target_model="dcim.virtualchassis"
        # should map to /technology/platforms/stack/members, not /inventory/devices
        if transform_map.source_model == "device" and transform_map.target_model:
            virtualchassis_ct = ContentType.objects.using(
                schema_editor.connection.alias
            ).get(app_label="dcim", model="virtualchassis")
            if transform_map.target_model_id == virtualchassis_ct.id:
                endpoint_value = "/technology/platforms/stack/members"

        # If no mapping exists, use fallback and mark the name for manual fix
        if not endpoint_value:
            if not transform_map.name.startswith("[NEEDS CORRECTION]"):
                transform_map.name = f"[NEEDS CORRECTION - Unknown source_model: {transform_map.source_model}] {transform_map.name}"
            transform_map.source_endpoint = fallback_endpoint
            transform_map.save()
            continue

        # Try to get the mapped endpoint
        try:
            endpoint = IPFabricEndpoint.objects.using(
                schema_editor.connection.alias
            ).get(endpoint=endpoint_value)
            transform_map.source_endpoint = endpoint
        except IPFabricEndpoint.DoesNotExist:
            # Use fallback endpoint and mark the name
            if not transform_map.name.startswith("[NEEDS CORRECTION]"):
                transform_map.name = f"[NEEDS CORRECTION - Expected endpoint '{endpoint_value}' not found] {transform_map.name}"
            transform_map.source_endpoint = fallback_endpoint
        transform_map.save()


def migrate_endpoint_to_source_model(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Reverse migration: migrate IPFabricTransformMap source_endpoint data back to source_model."""
    IPFabricTransformMap = apps.get_model("ipfabric_netbox", "IPFabricTransformMap")
    IPFabricEndpoint = apps.get_model("ipfabric_netbox", "IPFabricEndpoint")
    ContentType = apps.get_model("contenttypes", "ContentType")

    endpoint_to_source_model = {
        "/inventory/sites/overview": "site",
        "/inventory/devices": "device",
        "/technology/platforms/stack/members": "virtualchassis",
        "/inventory/interfaces": "interface",
        "/inventory/part-numbers": "part_number",
        "/technology/vlans/site-summary": "vlan",
        "/technology/routing/vrf/detail": "vrf",
        "/technology/networks/managed-networks": "prefix",
        "/technology/addressing/managed-ip/ipv4": "ipaddress",
    }

    for transform_map in IPFabricTransformMap.objects.using(
        schema_editor.connection.alias
    ).all():
        if not transform_map.source_endpoint:
            continue

        # Get the endpoint and map it back to source_model
        try:
            endpoint = IPFabricEndpoint.objects.using(
                schema_editor.connection.alias
            ).get(pk=transform_map.source_endpoint_id)
            source_model_value = endpoint_to_source_model.get(endpoint.endpoint)

            # Special case: source_endpoint="/technology/platforms/stack/members" with
            # target_model="dcim.virtualchassis" should map back to source_model="device"
            if (
                endpoint.endpoint == "/technology/platforms/stack/members"
                and transform_map.target_model
            ):
                virtualchassis_ct = ContentType.objects.using(
                    schema_editor.connection.alias
                ).get(app_label="dcim", model="virtualchassis")
                if transform_map.target_model_id == virtualchassis_ct.id:
                    source_model_value = "device"

            transform_map.source_model = source_model_value
        except IPFabricEndpoint.DoesNotExist:
            pass

        # Restore original name by removing correction markers
        if transform_map.name.startswith("[NEEDS CORRECTION"):
            # Extract original name after the correction marker
            import re

            match = re.match(r"\[NEEDS CORRECTION[^]]*]\s*(.+)", transform_map.name)
            if match:
                transform_map.name = match.group(1)
        transform_map.save()


transform_keys = {
    "site": "dcim.site",
    "vlan": "ipam.vlan",
    "manufacturer": "dcim.manufacturer",
    "vrf": "ipam.vrf",
    "platform": "dcim.platform",
    "devicerole": "dcim.devicerole",
    "devicetype": "dcim.devicetype",
    "prefix": "ipam.prefix",
    "device": "dcim.device",
    "virtualchassis": "dcim.virtualchassis",
    "ipaddress": "ipam.ipaddress",
    "interface": "dcim.interface",
    "inventoryitem": "dcim.inventoryitem",
    "macaddress": "dcim.macaddress",
}


def correct_sync_parameters(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Correct existing sync parameters to match new structure."""
    IPFabricSync = apps.get_model("ipfabric_netbox", "IPFabricSync")
    for sync in IPFabricSync.objects.using(schema_editor.connection.alias).all():
        for old_key, new_key in transform_keys.items():
            if old_key not in sync.parameters:
                continue
            sync.parameters[new_key] = sync.parameters.pop(old_key)
            sync.save()


def return_sync_parameters(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Reverse the key renaming in sync parameters to restore original structure."""
    IPFabricSync = apps.get_model("ipfabric_netbox", "IPFabricSync")
    reverse_transform_keys = {v: k for k, v in transform_keys.items()}
    for sync in IPFabricSync.objects.using(schema_editor.connection.alias).all():
        for new_key, old_key in reverse_transform_keys.items():
            if new_key not in sync.parameters:
                continue
            sync.parameters[old_key] = sync.parameters.pop(new_key)
            sync.save()


def assign_parent_transform_maps(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Assign parent relationships to transform maps based on transform_map.json data."""

    IPFabricTransformMap = apps.get_model("ipfabric_netbox", "IPFabricTransformMap")
    ContentType = apps.get_model("contenttypes", "ContentType")
    transform_map_data = get_transform_map()

    # Process each transform map from JSON
    for tm_data in transform_map_data:
        data = tm_data.get("data", {})

        parents_value = data.get("parents")
        if not parents_value:
            continue

        # Handle both string (single parent) and array (multiple parents) formats
        if isinstance(parents_value, str):
            parents_value = [parents_value]

        try:
            # Find the transform map (without group) matching this target model
            target_model_data = data.get("target_model", {})
            target_model = ContentType.objects.using(
                schema_editor.connection.alias
            ).get(
                app_label=target_model_data.get("app_label"),
                model=target_model_data.get("model"),
            )
            transform_map = (
                IPFabricTransformMap.objects.using(schema_editor.connection.alias)
                .filter(group__isnull=True, target_model=target_model)
                .first()
            )
            if not transform_map:
                continue

            # Add each parent to the M2M relationship
            for parent_str in parents_value:
                parent_app_label, parent_model = parent_str.split(".")

                try:
                    parent_content_type = ContentType.objects.using(
                        schema_editor.connection.alias
                    ).get(app_label=parent_app_label, model=parent_model)
                    parent_transform_map = (
                        IPFabricTransformMap.objects.using(
                            schema_editor.connection.alias
                        )
                        .filter(group__isnull=True, target_model=parent_content_type)
                        .first()
                    )
                    if parent_transform_map:
                        transform_map.parents.add(parent_transform_map)

                except ContentType.DoesNotExist:
                    continue

        except ContentType.DoesNotExist:
            continue


class Migration(migrations.Migration):
    dependencies = [
        ("extras", "0132_configcontextprofile"),
        ("ipfabric_netbox", "0022_prepare_for_filters"),
    ]

    operations = [
        migrations.RunPython(
            prepare_endpoints,
            migrations.RunPython.noop,
        ),
        migrations.RunPython(
            prepare_filters,
            migrations.RunPython.noop,
        ),
        migrations.RunPython(
            assign_filters_to_syncs,
            migrations.RunPython.noop,
        ),
        migrations.RunPython(
            migrate_source_model_to_endpoint,
            migrate_endpoint_to_source_model,
        ),
        migrations.RunPython(
            correct_sync_parameters,
            return_sync_parameters,
        ),
        migrations.RunPython(
            assign_parent_transform_maps,
            migrations.RunPython.noop,
        ),
    ]



# Generated by Django 5.2.5 on 2025-12-02 12:06
import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("extras", "0132_configcontextprofile"),
        ("ipfabric_netbox", "0023_populate_filters_data"),
    ]

    operations = [
        # The new endpoint is populated, so we can make the field non-nullable
        migrations.AlterField(
            model_name="ipfabrictransformmap",
            name="source_endpoint",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="transform_maps",
                to="ipfabric_netbox.ipfabricendpoint",
            ),
        ),
        # Remove the old source_model field as it's been replaced by source_endpoint
        migrations.RemoveField(
            model_name="ipfabrictransformmap",
            name="source_model",
        ),
    ]



from django.db.models.signals import post_delete
from netbox.plugins import PluginConfig


class NetboxIPFabricConfig(PluginConfig):
    name = "ipfabric_netbox"
    verbose_name = "NetBox IP Fabric SoT Plugin"
    description = "Sync IP Fabric into NetBox"
    version = "5.1.0b2"
    base_url = "ipfabric"
    min_version = "4.5.0"

    def ready(self):
        super().ready()

        from ipfabric_netbox.signals import remove_group_from_syncs

        post_delete.connect(
            remove_group_from_syncs,
            sender="ipfabric_netbox.IPFabricTransformMapGroup",
            dispatch_uid="remove_group_from_syncs",
        )


config = NetboxIPFabricConfig



import contextlib
import re
from typing import TYPE_CHECKING

import django.db.models.deletion
from django.db import migrations
from django.db import models

from ipfabric_netbox.utilities.transform_map import build_transform_maps
from ipfabric_netbox.utilities.transform_map import get_transform_map


if TYPE_CHECKING:
    from django.apps import apps
    from django.db.backends.base.schema import BaseDatabaseSchemaEditor
    from ipfabric_netbox.models import IPFabricTransformField, IPFabricTransformMap


def get_current_transform_map(
    apps: "apps",
    target_model: str,
    app_label: str = "dcim",
    source_model: str = "interface",
) -> "IPFabricTransformMap":
    return apps.get_model("ipfabric_netbox", "IPFabricTransformMap").objects.get(
        source_model=source_model,
        target_model=apps.get_model("contenttypes", "ContentType").objects.get(
            app_label=app_label,
            model=target_model,
        ),
    )


def get_current_map_field(apps: "apps", target_model: str) -> "IPFabricTransformField":
    """Finds the current TransformMapField for MAC address so we keep the template.
    We need to do this because some customer might have altered it."""
    return apps.get_model("ipfabric_netbox", "IPFabricTransformField").objects.get(
        source_field="mac",
        target_field="mac_address",
        transform_map=get_current_transform_map(apps, target_model),
    )


def create_transform_map(apps: "apps", schema_editor: "BaseDatabaseSchemaEditor"):
    """Create transform map structure for MAC Address (new in NetBox v4.2).
    Keeps current template for MAC Address if it exists.

    Also fixes interface.duplex not being populated."""

    with contextlib.suppress(
        apps.get_model("ipfabric_netbox", "IPFabricTransformMap").DoesNotExist
    ):
        # The map will already be present if we're using newer transform map.
        # This means we don't need to run the migration since it got created with 0008_prepare_transform_maps migration
        get_current_transform_map(apps, "macaddress")
        return

    # region - MAC Address
    current_mac_field = get_current_map_field(apps, "interface")
    transform_map_data = get_transform_map()

    mac_address_transform_map = None
    # Get MAC Address transform map defined in transform_map.json
    for i, transform_map in enumerate(transform_map_data[:]):
        if transform_map["data"]["name"] == "MAC Address Transform Map":
            mac_address_transform_map = [transform_map]
    # Replace the template only the current field is there
    if current_mac_field:
        for field_map in mac_address_transform_map[0]["field_maps"]:
            if field_map["source_field"] == "mac":
                field_map["template"] = current_mac_field.template

    build_transform_maps(mac_address_transform_map)
    current_mac_field.delete()
    # endregion

    # region - duplex
    # Create duplex transform field for Interface. Ignore if it already exists
    IPFabricTransformField = apps.get_model("ipfabric_netbox", "IPFabricTransformField")
    field_data = {
        "source_field": "duplex",
        "target_field": "duplex",
        "transform_map": get_current_transform_map(apps, "interface"),
    }
    try:
        IPFabricTransformField.objects.get(**field_data)
    except IPFabricTransformField.DoesNotExist:
        IPFabricTransformField(**field_data).save()
    # endregion

    # region - Prefix site
    # Prefix.site has changed to Prefix.scope relation (to allow site groups etc.)
    prefix_transform_map = get_current_transform_map(
        apps, source_model="prefix", app_label="ipam", target_model="prefix"
    )
    current_site_relationship = prefix_transform_map.relationship_maps.get(
        target_field="site"
    )
    IPFabricTransformField(
        source_field="siteName",
        target_field="scope_id",
        coalesce=True,
        template="{% if object.siteName is defined %}"
        + current_site_relationship.template
        + "{% else %}None{% endif %}",
        transform_map=prefix_transform_map,
    ).save()
    current_site_relationship.template = '{% if object.siteName is defined %}{{ contenttypes.ContentType.objects.get(app_label="dcim", model="site").pk }}{% else %}None{% endif %}'
    current_site_relationship.source_model = apps.get_model(
        "contenttypes", "ContentType"
    ).objects.get(app_label="contenttypes", model="contenttype")
    current_site_relationship.target_field = "scope_type"
    current_site_relationship.save()
    # endregion


def delete_transform_map(apps: "apps", schema_editor: "BaseDatabaseSchemaEditor"):
    current_mac_field = get_current_map_field(apps, "macaddress")

    # region - MAC Address
    apps.get_model("ipfabric_netbox", "IPFabricTransformField")(
        coalesce=False,
        source_field="mac",
        target_field="mac_address",
        template=current_mac_field.template,
        transform_map=get_current_transform_map(apps, "interface"),
    ).save()

    # Use CASCADE to delete the transform fields and relationships
    current_mac_field.transform_map.delete()
    # endregion

    # region - duplex
    # Delete new duplex field
    apps.get_model("ipfabric_netbox", "IPFabricTransformField").objects.get(
        source_field="duplex",
        target_field="duplex",
        transform_map=get_current_transform_map(apps, "interface"),
    ).delete()
    # endregion

    # region - Prefix site
    prefix_transform_map = get_current_transform_map(
        apps, source_model="prefix", app_label="ipam", target_model="prefix"
    )
    current_site_relationship = prefix_transform_map.relationship_maps.get(
        target_field="scope_type"
    )
    current_site_field = prefix_transform_map.field_maps.get(source_field="siteName")
    match = re.search(r"is defined %}(.*){% else %}None", current_site_field.template)
    current_site_field.delete()
    if match:
        current_site_relationship.template = match.group(1)
    else:
        current_site_relationship.template = "{% set SLUG = object.siteName | slugify %}{{ dcim.Site.objects.filter(slug=SLUG).first().pk }}"
    current_site_relationship.source_model = apps.get_model(
        "contenttypes", "ContentType"
    ).objects.get(app_label="dcim", model="site")
    current_site_relationship.target_field = "site"
    current_site_relationship.save()
    # endregion


def set_macaddress_sync_param(apps: "apps", schema_editor: "BaseDatabaseSchemaEditor"):
    IPFabricSync = apps.get_model("ipfabric_netbox", "IPFabricSync")
    for sync in IPFabricSync.objects.all():
        if sync.parameters.get("interface"):
            sync.parameters["macaddress"] = True
        else:
            sync.parameters["macaddress"] = False
        sync.save()


def remove_macaddress_sync_param(
    apps: "apps", schema_editor: "BaseDatabaseSchemaEditor"
):
    IPFabricSync = apps.get_model("ipfabric_netbox", "IPFabricSync")
    for sync in IPFabricSync.objects.all():
        sync.parameters.pop("macaddress", None)
        sync.save()


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("ipfabric_netbox", "0008_prepare_transform_maps"),
    ]

    operations = [
        migrations.AlterField(
            model_name="ipfabrictransformmap",
            name="target_model",
            field=models.ForeignKey(
                limit_choices_to=models.Q(
                    models.Q(
                        models.Q(("app_label", "dcim"), ("model", "site")),
                        models.Q(("app_label", "dcim"), ("model", "manufacturer")),
                        models.Q(("app_label", "dcim"), ("model", "platform")),
                        models.Q(("app_label", "dcim"), ("model", "devicerole")),
                        models.Q(("app_label", "dcim"), ("model", "devicetype")),
                        models.Q(("app_label", "dcim"), ("model", "device")),
                        models.Q(("app_label", "dcim"), ("model", "virtualchassis")),
                        models.Q(("app_label", "dcim"), ("model", "interface")),
                        models.Q(("app_label", "dcim"), ("model", "macaddress")),
                        models.Q(("app_label", "ipam"), ("model", "vlan")),
                        models.Q(("app_label", "ipam"), ("model", "vrf")),
                        models.Q(("app_label", "ipam"), ("model", "prefix")),
                        models.Q(("app_label", "ipam"), ("model", "ipaddress")),
                        models.Q(
                            ("app_label", "contenttypes"), ("model", "contenttype")
                        ),
                        models.Q(("app_label", "tenancy"), ("model", "tenant")),
                        models.Q(("app_label", "dcim"), ("model", "inventoryitem")),
                        _connector="OR",
                    )
                ),
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="contenttypes.contenttype",
            ),
        ),
        migrations.RunPython(set_macaddress_sync_param, remove_macaddress_sync_param),
        migrations.RunPython(create_transform_map, delete_transform_map),
    ]



import importlib.resources
import json

from netbox.settings import django_apps

from ..choices import IPFabricSourceTypeChoices

# region Filter Expression


def get_filter_expression_test_candidates(expression_instance):
    """
    Get potential LOCAL sources and endpoints from filters associated with a filter expression.

    Args:
        expression_instance: An IPFabricFilterExpression instance

    Returns:
        tuple: (sources_set, endpoints_set) - Sets of IPFabricSource and IPFabricEndpoint objects
    """
    if not expression_instance or not expression_instance.pk:
        return set(), set()

    sources = set()
    endpoints = set()

    # Query all filters that use this expression
    for filter_obj in expression_instance.filters.all():
        # Get LOCAL sources from syncs
        for sync in filter_obj.syncs.all():
            if (
                sync.snapshot_data
                and sync.snapshot_data.source
                and sync.snapshot_data.source.type == IPFabricSourceTypeChoices.LOCAL
            ):
                sources.add(sync.snapshot_data.source)

        # Get endpoints
        for endpoint in filter_obj.endpoints.all():
            endpoints.add(endpoint)

    return sources, endpoints


# endregion Filter Expression


# region Filter Creation

# These functions are used in the migration file to prepare the filters
# Because of this we have to use historical models
# see https://docs.djangoproject.com/en/5.1/topics/migrations/#historical-models


def build_filters(data, apps: django_apps = None, db_alias: str = "default") -> None:
    apps = apps or django_apps
    IPFabricEndpoint = apps.get_model("ipfabric_netbox", "IPFabricEndpoint")
    IPFabricFilter = apps.get_model("ipfabric_netbox", "IPFabricFilter")
    IPFabricFilterExpression = apps.get_model(
        "ipfabric_netbox", "IPFabricFilterExpression"
    )
    for expressions_data in data["expressions"]:
        IPFabricFilterExpression.objects.using(db_alias).create(**expressions_data)
    for filter_data in data["filters"]:
        filter_data["filter_type"] = "and"
        endpoints = [
            IPFabricEndpoint.objects.filter(endpoint=endpoint).first()
            for endpoint in filter_data.pop("endpoints", [])
        ]
        expressions = [
            IPFabricFilterExpression.objects.filter(name=expr).first()
            for expr in filter_data.pop("expressions", [])
        ]
        _filter = IPFabricFilter.objects.using(db_alias).create(**filter_data)
        _filter.endpoints.set(endpoints)
        _filter.expressions.set(expressions)


def get_filter_data() -> dict:
    for data_file in importlib.resources.files("ipfabric_netbox.data").iterdir():
        if data_file.name != "filters.json":
            continue
        with open(data_file, "rb") as data_file:
            return json.load(data_file)
    raise FileNotFoundError("'filters.json' not found in installed package")


# endregion Filter Creation



from django.db import migrations

from ipfabric_netbox.utilities.transform_map import do_change
from ipfabric_netbox.utilities.transform_map import RelationshipRecord
from ipfabric_netbox.utilities.transform_map import TransformMapRecord


TRANSFORM_MAP_CHANGES = (
    TransformMapRecord(
        source_endpoint="/technology/platforms/stack/members",
        target_model="dcim.virtualchassis",
        relationships=(
            RelationshipRecord(
                source_model="dcim.device",
                target_field="master",
                old_template="{% set DEVICE = dcim.Device.objects.filter(serial=object.sn).first() %}{% if DEVICE %}{{ DEVICE.pk }}{% endif %}",
                new_template="{% set DEVICE = dcim.Device.objects.filter(serial=object.sn).first() %}{% if DEVICE %}{% if object.master is defined and object.master %}{% set VC_NAME = object.master %}{% else %}{% set VC_NAME = object.hostname %}{% endif %}{% set VIRTUAL_CHASSIS = dcim.VirtualChassis.objects.filter(name=VC_NAME).first() %}{% if VIRTUAL_CHASSIS and DEVICE.virtual_chassis == VIRTUAL_CHASSIS %}{{ DEVICE.pk }}{% endif %}{% endif %}",
            ),
        ),
    ),
)


def forward_transform_maps_change(apps, schema_editor):
    """Replace old template with updated version."""
    do_change(apps, schema_editor, changes=TRANSFORM_MAP_CHANGES, forward=True)


def revert_transform_maps_change(apps, schema_editor):
    """Revert template back to the previous exact template."""
    do_change(apps, schema_editor, changes=TRANSFORM_MAP_CHANGES, forward=False)


class Migration(migrations.Migration):
    dependencies = [
        ("ipfabric_netbox", "0026_update_platform_templates"),
    ]

    operations = [
        migrations.RunPython(
            forward_transform_maps_change,
            revert_transform_maps_change,
        ),
    ]



import django_filters
from core.choices import ObjectChangeActionChoices
from django.contrib.contenttypes.models import ContentType
from django.db.models import JSONField
from django.db.models import Q
from django.utils.translation import gettext as _
from netbox.filtersets import BaseFilterSet
from netbox.filtersets import ChangeLoggedModelFilterSet
from netbox.filtersets import NetBoxModelFilterSet
from netbox_branching.models import ChangeDiff

from .choices import IPFabricSourceStatusChoices
from .choices import IPFabricSyncStatusChoices
from .models import IPFabricData
from .models import IPFabricEndpoint
from .models import IPFabricFilter
from .models import IPFabricFilterExpression
from .models import IPFabricIngestion
from .models import IPFabricIngestionIssue
from .models import IPFabricRelationshipField
from .models import IPFabricSnapshot
from .models import IPFabricSource
from .models import IPFabricSync
from .models import IPFabricTransformField
from .models import IPFabricTransformMap
from .models import IPFabricTransformMapGroup


class IPFabricIngestionChangeFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(method="search")
    action = django_filters.MultipleChoiceFilter(choices=ObjectChangeActionChoices)

    class Meta:
        model = ChangeDiff
        fields = ["branch", "action", "object_type"]

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(current__icontains=value)
            | Q(modified__icontains=value)
            | Q(original__icontains=value)
            | Q(action__icontains=value)
            | Q(object_type__model__icontains=value)
        )


class IPFabricIngestionIssueFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(method="search")

    class Meta:
        model = IPFabricIngestionIssue
        fields = [
            "model",
            "timestamp",
            "raw_data",
            "coalesce_fields",
            "defaults",
            "exception",
            "message",
        ]

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(model__icontains=value)
            | Q(timestamp__icontains=value)
            | Q(raw_data__icontains=value)
            | Q(coalesce_fields__icontains=value)
            | Q(defaults__icontains=value)
            | Q(exception__icontains=value)
            | Q(message__icontains=value)
        )


class IPFabricDataFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(method="search")

    class Meta:
        model = IPFabricData
        fields = ["snapshot_data"]

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        # Search by snapshot name since snapshot_data is a ForeignKey
        return queryset.filter(Q(snapshot_data__name__icontains=value))


class IPFabricSnapshotFilterSet(ChangeLoggedModelFilterSet):
    q = django_filters.CharFilter(method="search")
    source_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricSource.objects.all(),
        label=_("Source (ID)"),
    )
    source = django_filters.ModelMultipleChoiceFilter(
        field_name="source__name",
        queryset=IPFabricSource.objects.all(),
        to_field_name="name",
        label=_("Source (name)"),
    )
    snapshot_id = django_filters.CharFilter(
        label=_("Snapshot ID"), lookup_expr="icontains"
    )

    class Meta:
        model = IPFabricSnapshot
        fields = ("id", "name", "status", "snapshot_id")

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(Q(name__icontains=value))


class IPFabricSourceFilterSet(NetBoxModelFilterSet):
    status = django_filters.MultipleChoiceFilter(
        choices=IPFabricSourceStatusChoices, null_value=None
    )

    class Meta:
        model = IPFabricSource
        fields = ("id", "name")

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value)
            | Q(description__icontains=value)
            | Q(comments__icontains=value)
        )


class IPFabricSyncFilterSet(ChangeLoggedModelFilterSet):
    q = django_filters.CharFilter(method="search")
    name = django_filters.CharFilter()
    snapshot_data_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricSnapshot.objects.all(),
        label=_("Snapshot (ID)"),
    )
    snapshot_data = django_filters.ModelMultipleChoiceFilter(
        field_name="snapshot_data__name",
        queryset=IPFabricSnapshot.objects.all(),
        to_field_name="name",
        label=_("Snapshot (name)"),
    )
    status = django_filters.MultipleChoiceFilter(
        choices=IPFabricSyncStatusChoices, null_value=None
    )

    ipfabric_filter_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricFilter.objects.all(),
        field_name="filters",
        label=_("IP Fabric Filter (ID)"),
    )
    ipfabric_filter = django_filters.CharFilter(
        field_name="filters__name",
        lookup_expr="iexact",
        label=_("IP Fabric Filter (name)"),
        distinct=True,
    )
    ipfabric_filters = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricFilter.objects.all(),
        field_name="filters",
        label=_("IP Fabric Filters"),
    )

    class Meta:
        model = IPFabricSync
        fields = (
            "id",
            "name",
            "snapshot_data",
            "snapshot_data_id",
            "status",
            "auto_merge",
            "last_synced",
            "scheduled",
            "interval",
            "ipfabric_filter_id",
            "ipfabric_filter",
            "ipfabric_filters",
        )

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value) | Q(snapshot_data__name__icontains=value)
        )


class IPFabricIngestionFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(method="search")
    sync_id = django_filters.ModelMultipleChoiceFilter(
        field_name="sync",
        queryset=IPFabricSync.objects.all(),
        label=_("Sync (ID)"),
    )
    sync = django_filters.ModelMultipleChoiceFilter(
        field_name="sync__name",
        queryset=IPFabricSync.objects.all(),
        to_field_name="name",
        label=_("Sync (name)"),
    )

    class Meta:
        model = IPFabricIngestion
        fields = ("id", "branch", "sync")

    def search(self, queryset, branch, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(branch__name__icontains=value) | Q(sync__name__icontains=value)
        )


class IPFabricFilterFilterSet(NetBoxModelFilterSet):
    q = django_filters.CharFilter(method="search")

    sync_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricSync.objects.all(),
        field_name="syncs",
        label=_("Syncs (ID)"),
    )
    sync = django_filters.CharFilter(
        field_name="syncs__name",
        lookup_expr="iexact",
        label=_("Sync (name)"),
        distinct=True,
    )
    syncs = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricSync.objects.all(),
        field_name="syncs",
        label=_("Syncs (ID)"),
    )

    endpoint_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricEndpoint.objects.all(),
        field_name="endpoints",
        label=_("Endpoints (ID)"),
    )
    endpoint = django_filters.CharFilter(
        field_name="endpoints__name",
        lookup_expr="iexact",
        label=_("Endpoint (Name)"),
        distinct=True,
    )
    endpoint_path = django_filters.CharFilter(
        field_name="endpoints__endpoint",
        lookup_expr="iexact",
        label=_("Endpoint (Path)"),
        distinct=True,
    )
    endpoints = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricEndpoint.objects.all(),
        field_name="endpoints",
        label=_("Endpoints (ID)"),
    )

    expression_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricFilterExpression.objects.all(),
        field_name="expressions",
        label=_("Expression (ID)"),
    )
    expression = django_filters.CharFilter(
        field_name="expressions__name",
        lookup_expr="iexact",
        label=_("Expression (name)"),
        distinct=True,
    )
    expressions = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricFilterExpression.objects.all(),
        field_name="expressions",
        label=_("Expressions (ID)"),
    )

    class Meta:
        model = IPFabricFilter
        fields = (
            "id",
            "name",
            "description",
            "endpoint_id",
            "endpoint",
            "endpoint_path",
            "endpoints",
            "filter_type",
            "sync_id",
            "sync",
            "syncs",
            "expression_id",
            "expression",
            "expressions",
        )

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value) | Q(description__icontains=value)
        )


class IPFabricFilterExpressionFilterSet(NetBoxModelFilterSet):
    q = django_filters.CharFilter(method="search")

    ipfabric_filter_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricFilter.objects.all(),
        field_name="filters",
        label=_("Filter (ID)"),
    )
    ipfabric_filter = django_filters.CharFilter(
        field_name="filters__name",
        lookup_expr="iexact",
        label=_("Filters (name)"),
        distinct=True,
    )
    ipfabric_filters = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricFilter.objects.all(),
        field_name="filters",
        label=_("Filters (ID)"),
    )

    class Meta:
        model = IPFabricFilterExpression
        fields = (
            "id",
            "name",
            "description",
            "expression",
            "ipfabric_filter_id",
            "ipfabric_filter",
            "ipfabric_filters",
        )
        filter_overrides = {
            JSONField: {
                "filter_class": django_filters.CharFilter,
                "extra": lambda f: {"lookup_expr": "icontains"},
            },
        }

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value) | Q(description__icontains=value)
        )


class IPFabricEndpointFilterSet(NetBoxModelFilterSet):
    q = django_filters.CharFilter(method="search")

    endpoint = django_filters.CharFilter(
        field_name="endpoint",
        lookup_expr="iexact",
        label=_("Endpoint (Path)"),
    )
    ipfabric_filter_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricFilter.objects.all(),
        field_name="filters",
        label=_("IP Fabric Filter (ID)"),
    )
    ipfabric_filter = django_filters.CharFilter(
        field_name="filters__name",
        lookup_expr="iexact",
        label=_("IP Fabric Filter (name)"),
        distinct=True,
    )
    ipfabric_filters = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricFilter.objects.all(),
        field_name="filters",
        label=_("IP Fabric Filters"),
    )

    class Meta:
        model = IPFabricEndpoint
        fields = (
            "id",
            "name",
            "description",
            "endpoint",
            "ipfabric_filter_id",
            "ipfabric_filter",
            "ipfabric_filters",
        )

    def search(self, queryset, name, value):
        if not value.strip().rstrip("/"):
            return queryset
        return queryset.filter(
            Q(name__icontains=value)
            | Q(endpoint__icontains=value)
            | Q(description__icontains=value)
        )


class IPFabricTransformMapGroupFilterSet(NetBoxModelFilterSet):
    q = django_filters.CharFilter(method="search")

    class Meta:
        model = IPFabricTransformMapGroup
        fields = ("id", "name", "description")

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value) | Q(description__icontains=value)
        )


class IPFabricTransformMapFilterSet(NetBoxModelFilterSet):
    q = django_filters.CharFilter(method="search")
    group_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricTransformMapGroup.objects.all(),
        label=_("Transform Map Group (ID)"),
    )
    group = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricTransformMapGroup.objects.all(), label=_("Transform Map Group")
    )
    source_endpoint = django_filters.ModelChoiceFilter(
        queryset=IPFabricEndpoint.objects.all(), label=_("Source Endpoint")
    )
    source_endpoint_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricEndpoint.objects.all(),
        field_name="source_endpoint",
        label=_("Source Endpoint (ID)"),
    )
    source_endpoints = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricEndpoint.objects.all(),
        field_name="source_endpoint",
        label=_("Source Endpoints"),
    )

    class Meta:
        model = IPFabricTransformMap
        fields = (
            "id",
            "name",
            "group",
            "source_endpoint",
            "source_endpoint_id",
            "source_endpoints",
            "target_model",
        )

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(group__name__icontains=value) | Q(name__icontains=value)
        )


class IPFabricTransformFieldFilterSet(BaseFilterSet):
    transform_map = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricTransformMap.objects.all(), label=_("Transform Map")
    )
    transform_map_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricTransformMap.objects.all(),
        field_name="transform_map",
        label=_("Transform Map (ID)"),
    )
    source_field = django_filters.CharFilter(
        field_name="source_field",
        lookup_expr="exact",
        label=_("Source Field"),
    )
    target_field = django_filters.CharFilter(
        field_name="target_field",
        lookup_expr="exact",
        label=_("Target Field"),
    )
    coalesce = django_filters.BooleanFilter(
        field_name="coalesce",
        label=_("Coalesce"),
    )

    class Meta:
        model = IPFabricTransformField
        fields = (
            "id",
            "transform_map",
            "transform_map_id",
            "source_field",
            "target_field",
            "coalesce",
            "template",
        )


class IPFabricRelationshipFieldFilterSet(BaseFilterSet):
    transform_map = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricTransformMap.objects.all(), label=_("Transform Map")
    )
    transform_map_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricTransformMap.objects.all(),
        field_name="transform_map",
        label=_("Transform Map (ID)"),
    )
    source_model = django_filters.ModelChoiceFilter(
        queryset=ContentType.objects.all(),
        field_name="source_model",
        label=_("Source Model"),
    )
    target_field = django_filters.CharFilter(
        field_name="target_field",
        lookup_expr="exact",
        label=_("Target Field"),
    )
    coalesce = django_filters.BooleanFilter(
        field_name="coalesce",
        label=_("Coalesce"),
    )

    class Meta:
        model = IPFabricRelationshipField
        fields = (
            "id",
            "transform_map",
            "transform_map_id",
            "source_model",
            "target_field",
            "coalesce",
            "template",
        )



"""
NIM-18852.
Update to truncate name and description fields because they can
exceed DB field limits.
"""
from django.db import migrations

# Keep the original template here rather than loading it from transform_map.json
# so our revert won’t break if that template ever changes.
OLD_TEMPLATE = (
    "{% if object.name is not none %}"
    "{{ object.name }}"
    "{% elif object.dscr is not none %}"
    "{{ object.dscr}}"
    "{% else %}Default Name{% endif %}"
)

NEW_TEMPLATE = (
    "{% if object.name is not none %}"
    "{{ object.name | string | truncate(64, True) }}"
    "{% elif object.dscr is not none %}"
    "{{ object.dscr | string | truncate(64, True) }}"
    "{% else %}Default Name{% endif %}"
)


def apply_truncate_template(apps, schema_editor):
    """Replace old template with truncated version on all matching transform fields."""
    TransformField = apps.get_model("ipfabric_netbox", "IPFabricTransformField")
    TransformField.objects.filter(
        template=OLD_TEMPLATE,
        source_field="name",
        target_field="name",
    ).update(template=NEW_TEMPLATE)


def revert_truncate_template(apps, schema_editor):
    """Revert truncated template back to the original exact template."""
    TransformField = apps.get_model("ipfabric_netbox", "IPFabricTransformField")
    TransformField.objects.filter(
        template=NEW_TEMPLATE,
        source_field="name",
        target_field="name",
    ).update(template=OLD_TEMPLATE)


class Migration(migrations.Migration):
    dependencies = [
        ("ipfabric_netbox", "0010_remove_uuid_from_get_or_create"),
    ]

    operations = [
        migrations.RunPython(
            apply_truncate_template,
            revert_truncate_template,
        ),
    ]



from typing import TYPE_CHECKING

from django.db import IntegrityError
from django.db import migrations
from django.utils.text import slugify

from ipfabric_netbox.utilities.transform_map import do_change
from ipfabric_netbox.utilities.transform_map import FieldRecord
from ipfabric_netbox.utilities.transform_map import RelationshipRecord
from ipfabric_netbox.utilities.transform_map import TransformMapRecord

if TYPE_CHECKING:
    from django.apps import apps as apps_type
    from django.db.backends.base.schema import BaseDatabaseSchemaEditor

TRANSFORM_MAP_CHANGES = (
    # Platform Transform Map changes
    TransformMapRecord(
        source_endpoint="/inventory/devices",
        target_model="dcim.platform",
        fields=(
            FieldRecord(
                source_field="vendor",
                target_field="slug",
                old_template="{{ object.vendor | slugify }}{%if object.family %}_{{ object.family | slugify }}{% endif %}",
                new_template="{% if object.family %}{{ object.family | slugify }}{% else %}{{ object.vendor | slugify }}{% endif %}",
            ),
            FieldRecord(
                source_field="family",
                target_field="name",
                old_template="{%if object.family %}{{ object.family | slugify }}{% else %}{{ object.vendor }}{% endif %}",
                new_template="{% if object.family %}{{ object.family }}{% else %}{{ object.vendor }}{% endif %}",
            ),
        ),
        relationships=(
            RelationshipRecord(
                source_model="dcim.manufacturer",
                target_field="manufacturer",
                coalesce=True,
            ),
        ),
    ),
    # Device Transform Map changes - platform relationship
    TransformMapRecord(
        source_endpoint="/inventory/devices",
        target_model="dcim.device",
        relationships=(
            RelationshipRecord(
                source_model="dcim.platform",
                target_field="platform",
                old_template='{% set SLUG = object.vendor | slugify %}{% if object.family %}{% set SLUG = SLUG ~ "_" ~ object.family | slugify %}{% endif %}\n{{ dcim.Platform.objects.get(slug=SLUG).pk }}',
                new_template="{% set MANUFACTURER_SLUG = object.vendor | slugify %}{% if object.family %}{% set SLUG = object.family | slugify %}{% else %}{% set SLUG = object.vendor | slugify %}{% endif %}{{ dcim.Platform.objects.get(slug=SLUG, manufacturer__slug=MANUFACTURER_SLUG).pk }}",
            ),
        ),
    ),
)


def forward_transform_maps_change(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Replace old template with updated version."""
    do_change(apps, schema_editor, changes=TRANSFORM_MAP_CHANGES, forward=True)


def revert_transform_maps_change(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """Revert template back to the previous exact template."""
    do_change(apps, schema_editor, changes=TRANSFORM_MAP_CHANGES, forward=False)


def cleanup_platform_slugs_forward(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
):
    """
    Migrate existing Platform records by regenerating slug from name field.

    Always regenerates slug from the name field using slugify().
    New format: family OR vendor (whatever is in the name field)

    Examples:
    - name="IOS" -> slug="ios"
    - name="JunOS" -> slug="junos"
    - name="Palo Alto" -> slug="palo-alto"
    - name="Cisco" -> slug="cisco"

    Handles conflicts gracefully by skipping platforms that would create duplicates.
    """

    Platform = apps.get_model("dcim", "Platform")

    skipped_platforms = []
    updated_count = 0
    skipped_no_change = 0

    platforms = Platform.objects.using(schema_editor.connection.alias).all()

    for platform in platforms:
        old_slug = platform.slug

        # Always regenerate slug from name field
        # The name field contains either family or vendor (based on template logic)
        new_slug = slugify(platform.name)

        # If the slug is already correct, no change needed
        if old_slug == new_slug:
            skipped_no_change += 1
            continue

        try:
            # Update the slug
            platform.slug = new_slug
            platform.save(using=schema_editor.connection.alias)
            updated_count += 1
        except IntegrityError:
            # Slug already exists, skip this platform
            skipped_platforms.append(
                {
                    "old_slug": old_slug,
                    "attempted_new_slug": new_slug,
                    "name": platform.name,
                }
            )

    if skipped_platforms:
        print(
            f"Platform slug cleanup: Updated {updated_count} platforms, "
            f"skipped {len(skipped_platforms)} due to conflicts, "
            f"{skipped_no_change} unchanged: "
        )
        for skipped in skipped_platforms:
            print(
                f" - '{skipped['old_slug']}' -> '{skipped['attempted_new_slug']}' "
                f"(name: {skipped['name']}) - slug already exists"
            )
    else:
        print(
            f"Platform slug cleanup: Updated {updated_count} platforms successfully, "
            f"{skipped_no_change} unchanged"
        )


class Migration(migrations.Migration):
    dependencies = [
        ("ipfabric_netbox", "0025_add_vss_chassis_endpoint"),
    ]

    operations = [
        migrations.RunPython(
            forward_transform_maps_change,
            revert_transform_maps_change,
        ),
        migrations.RunPython(
            cleanup_platform_slugs_forward,
            # We cannot know which platforms were changed, so no-op on reverse
            migrations.RunPython.noop,
        ),
    ]



from datetime import timedelta
from typing import TYPE_CHECKING

from django.db import migrations
from utilities.datetime import local_now

if TYPE_CHECKING:
    from django.apps import apps as apps_type
    from django.db.backends.base.schema import BaseDatabaseSchemaEditor


def clean_scheduled_jobs(
    apps: "apps_type", schema_editor: "BaseDatabaseSchemaEditor"
) -> None:
    Job = apps.get_model("core", "Job")
    IPFabricSync = apps.get_model("ipfabric_netbox", "IPFabricSync")
    ObjectType = apps.get_model("core", "ObjectType")

    for sync in IPFabricSync.objects.all():
        try:
            scheduled_jobs = Job.objects.filter(
                object_id=sync.id,
                object_type=ObjectType.objects.get_for_model(sync),
                scheduled__isnull=False,
            ).order_by("scheduled")
            if not scheduled_jobs.exists():
                continue
            if not sync.scheduled:
                # Delete all scheduled jobs if the sync is not scheduled
                scheduled_jobs.delete()
                continue
            if scheduled_jobs.count() == 1:
                # Only one scheduled job exists, let's update scheduled time on the sync object
                # This does not create a new job since sync is a Faked object in migration
                sync.scheduled = scheduled_jobs.first().scheduled
                sync.full_clean()
                sync.save()
                continue
            # More than one scheduled job exists
            # Find the one that is closest to scheduled + N * interval
            interval = timedelta(minutes=sync.interval)
            elapsed = local_now() - sync.scheduled
            intervals_passed = (elapsed // interval) + 1
            closest_future_scheduled = sync.scheduled + intervals_passed * interval
            closest_job = min(
                scheduled_jobs,
                key=lambda job: abs(job.scheduled - closest_future_scheduled),
            )
            for job in scheduled_jobs:
                if job != closest_job:
                    job.delete()
            sync.scheduled = closest_job.scheduled
            sync.full_clean()
            sync.save()
        except Exception as err:
            # Always be safe inside a migration
            print(f"Error cleaning scheduled jobs for IPFabricSync {sync.id}: {err}")


class Migration(migrations.Migration):
    dependencies = [
        (
            "ipfabric_netbox",
            "0019_alter_ipfabrictransformmap_options_and_more",
        ),
    ]

    operations = [
        migrations.RunPython(clean_scheduled_jobs, migrations.RunPython.noop),
    ]
